from typing import Any, BinaryIO, List, Optional, Tuple
from pathlib import Path
import struct
import win32api
import click


# Documentation on struct specifications and their use can be found here:
# https://web.archive.org/web/20160531004250/https://msdn.microsoft.com/en-us/library/ms997538.aspx
ICONDIRHEADER = (("idReserved", "idType", "idCount"), "hhh")
ICONDIRENTRY = (
    (
        "bWidth",
        "bHeight",
        "bColorCount",
        "bReserved",
        "wPlanes",
        "wBitCount",
        "dwBytesInRes",
        "dwImageOffset",
    ),
    "bbbbhhii",
)
GRPICONDIRENTRY = (
    (
        "bWidth",
        "bHeight",
        "bColorCount",
        "bReserved",
        "wPlanes",
        "wBitCount",
        "dwBytesInRes",
        "nID",
    ),
    "bbbbhhih",
)
RT_ICON = 3
RT_GROUP_ICON = 14


class DataStruct(object):
    """General class for handling data structures within a file"""

    def __init__(
        self,
        dtype: Tuple[Tuple[str, ...], str],
        input_stream: Optional[BinaryIO] = None,
    ):
        self.__dict__["_field_names"] = dtype[0]
        self._data_types = dtype[1]
        assert len(self._field_names) == len(self._data_types)
        self._indices = {}
        for i, name in enumerate(self._field_names):
            self._indices[name] = i
        self._size = struct.calcsize(self._data_types)
        self._data = list(
            struct.unpack(
                self._data_types,
                bytearray(self._size)
                if input_stream is None
                else input_stream.read(self._size),
            )
        )

    def __getattr__(self, name: str) -> Any:
        if name in self._field_names:
            return self._data[self._indices[name]]
        return self.__dict__[name]

    def __setattr__(self, name: str, value: Any):
        if name in self._field_names:
            self._data[self._indices[name]] = value
        else:
            self.__dict__[name] = value

    def get_data(self) -> bytes:
        return struct.pack(self._data_types, *self._data)

    def copy(self, data_struct: "DataStruct"):
        for field_name in data_struct._field_names:
            if field_name in self._field_names:
                setattr(self, field_name, getattr(data_struct, field_name))


class Icon(object):
    """Class to extract the relevant data from a .ico file"""

    def __init__(self, file_name: Path):
        with open(str(file_name), "rb") as f:
            self._header = DataStruct(dtype=ICONDIRHEADER, input_stream=f)
            self._dir_entries = [
                DataStruct(dtype=ICONDIRENTRY, input_stream=f)
                for _ in range(self._header.idCount)
            ]
            self._icon_data: List[bytes] = []
            for entry in self._dir_entries:
                f.seek(entry.dwImageOffset, 0)
                self._icon_data.append(f.read(entry.dwBytesInRes))

    def get_header_and_group_icon_dir_data(self) -> bytes:
        data = self._header.get_data()
        for i, dir_entry in enumerate(self._dir_entries):
            icon_dir_entry = DataStruct(dtype=GRPICONDIRENTRY)
            icon_dir_entry.copy(data_struct=dir_entry)
            icon_dir_entry.nID = i + 1
            data += icon_dir_entry.get_data()
        return data

    def get_icon_data(self) -> List[bytes]:
        return self._icon_data


def add_icon_to_exe(source_icon_file: Path, target_exe_file: Path):
    target_exe_file = target_exe_file.absolute().resolve()
    if not target_exe_file.is_file():
        raise FileNotFoundError(
            f"The target executable file could not be found or is not a valid file: {target_exe_file}"
        )
    source_icon_file = source_icon_file.absolute().resolve()
    if not source_icon_file.is_file():
        raise FileNotFoundError(
            f"The icon file could not be found or is not a valid file: {source_icon_file}"
        )
    icon = Icon(file_name=source_icon_file)
    ur_handle = win32api.BeginUpdateResource(str(target_exe_file), 0)
    win32api.UpdateResource(
        ur_handle, RT_GROUP_ICON, 0, icon.get_header_and_group_icon_dir_data()
    )
    for i, icon_data in enumerate(icon.get_icon_data()):
        win32api.UpdateResource(ur_handle, RT_ICON, i + 1, icon_data)
    win32api.EndUpdateResource(ur_handle, 0)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument(
    "target_exe_file",
    type=click.Path(dir_okay=False, file_okay=True, exists=True),
)
@click.argument(
    "source_icon_file",
    type=click.Path(dir_okay=False, file_okay=True, exists=True),
)
def cli(target_exe_file: str, source_icon_file: str):
    """
    Add a .ico file to an existing .exe file

    The .exe file gets modified in place
    """
    add_icon_to_exe(
        source_icon_file=Path(source_icon_file),
        target_exe_file=Path(target_exe_file),
    )


if __name__ == "__main__":
    cli()



from typing import Optional
from pathlib import Path
import click
from genexe.winicon import add_icon_to_exe


SCRIPT_DIR = Path(__file__).parent.absolute().resolve()
RESOURCES_DIR = SCRIPT_DIR / "res"
EXE_TEMPLATE_FILE = RESOURCES_DIR / "template"
MAX_CMD_LENGTH = 259
REPLACE_SIGNATURE = (b"X" * MAX_CMD_LENGTH) + b"1"


def generate_exe(
    target: Path,
    command: str,
    icon_file: Optional[Path] = None,
    show_console: bool = True,
):
    target = target.absolute().resolve()
    if target == EXE_TEMPLATE_FILE:
        raise RuntimeError(
            "Cannot overwrite the source EXE_TEMPLATE_FILE file! Pick a different target executable name."
        )
    with open(EXE_TEMPLATE_FILE, "rb") as f:
        data = f.read()
    if len(command) > MAX_CMD_LENGTH:
        print(
            "Warning encode string was longer than MAX_CMD_LENGTH characters and will be truncated..."
        )
        command = command[:MAX_CMD_LENGTH]
    else:
        command = command + "\0" * (MAX_CMD_LENGTH - len(command))
    assert len(command) == MAX_CMD_LENGTH
    msg = command + ("1" if show_console else "0")
    byte_encoded_string = msg.encode("ascii")
    data = data.replace(REPLACE_SIGNATURE, byte_encoded_string)
    print(f"Encode show_console : {show_console}")
    print(f"Adding following command to exe : `{command.strip()}`")
    print(f"Writing exe to: {target}")
    with open(target, "wb") as f:
        f.write(data)
    if icon_file is not None:
        icon_file = Path(icon_file).absolute().resolve()
        print(f"Adding icon to exe: {icon_file}")
        add_icon_to_exe(target_exe_file=target, source_icon_file=icon_file)
    print("Done.")


class ClickPath(click.Path):
    """A Click path argument that returns a pathlib.Path object instead of a str"""

    def convert(self, value, param, ctx):
        return Path(super().convert(value, param, ctx))


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument(
    "target", type=ClickPath(exists=False, file_okay=True, dir_okay=False)
)
@click.argument("command", type=str)
@click.option(
    "-i",
    "--icon-file",
    "icon_file",
    type=ClickPath(exists=True, file_okay=True, dir_okay=False),
    default=None,
    help="Optional icon file to embed in the executable. Default is to use no icon-file",
)
@click.option(
    "-hc",
    "--hide-console",
    "show_console",
    is_flag=True,
    default=True,
    help="Hide the executable console window. Default show the console window",
)
def cli(
    target: Path, command: str, icon_file: Optional[Path], show_console: bool
):
    """
    Generate an executable which will run the provided command on your Windows system.

    The command argument supports the `{EXE_DIR}` macro, which expands the path where the
    executable is executed from.

    Optionally you can provide an icon file for the executable.
    """

    generate_exe(
        target=target,
        command=command,
        icon_file=icon_file,
        show_console=show_console,
    )


if __name__ == "__main__":
    cli()



from os import path
from setuptools import setup, find_packages


VERSION = "0.2.1"

this_directory = path.abspath(path.dirname(__file__))

with open(
    path.join(this_directory, "README.md"), mode="r", encoding="utf-8"
) as fh:
    long_description = fh.read()

with open(
    path.join(this_directory, "requirements.txt"), mode="r", encoding="utf-8"
) as fh:
    requirements = [e.strip() for e in fh.readlines() if e.strip() != ""]

setup(
    name="gen-exe",
    version=VERSION,
    author="S.C. van de Leemput",
    author_email="silvandeleemput@gmail.com",
    install_requires=requirements,
    include_package_data=True,
    license="LICENSE",
    entry_points={
        "console_scripts": [
            "gen-exe=genexe.generate_exe:cli",
            "add-ico-to-exe=genexe.winicon:cli",
        ]
    },
    packages=find_packages(),
    description="A small utility which allows you to generate Windows executables "
    "that can run custom commands on your Windows system.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Operating System :: Microsoft :: Windows",
        "Intended Audience :: Developers",
    ],
)
